/* This example shows the usage of pipe system call.
 *
 * This program opens a pipe and forks a child.
 * Then the parent write into the pipe some lines 
 * of text, that are read by the child.  
 *
 * NOTE: the writing and reading to/from the pipe,
 * is accomplished using the pipe descriptor.
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#define MYAPP "pipeParChComm"
#define PARENT_MSG "[PARENT]"
#define CHILD_MSG  "[CHILD] "

#define MESSAGE "Hello LSO1!" 
#define NUM_TIMES 5

char *src="";

/* Write COUNT copies of MESSAGE to pipe file descriptor, pausing for a second
   between each. */
void writer (const char* message, int count, int fd)
{
  int len, n;

  len=strlen(message);
  for (; count > 0; --count) {
    printf("%s %s: [%d] writing message \"%s\" \n",MYAPP,src,count,message);

    /* Write the message to the stream, and send it off immediately. */
    n=write(fd,message,len);
    if (n!=len)
      fprintf(stderr,"%s %s: ERROR writing %d bytes to pipe (ret=%d)\n",MYAPP,src,len,n);

    /* Snooze a while. */
    sleep (1);
  }
}

/* Read random strings from the pipe as long as possible.  */
void reader (int fd)
{
  char buffer[1024];
  int count=1;
  int n;
  int more=1;

  /* Read until we get 0 bytes, that means the the write-end of pipe
     has been closed. 
     REMIND read() from a pipe is a blocking function in the case 
     there are no bytes to read AND the write-end is OPEN.
     So, the reader can detect the closing of the write-end 
     (which means normally that the writer has completed its 
     communication) when the read() returns 0.
  */
  while (more) {
    printf("%s %s: [%d] reading message from read-end of pipe\n",MYAPP,src,count);
    if ( (n = read(fd,buffer,sizeof(buffer))) > 0) {
      printf("%s %s: [%d] %d bytes read!\n",MYAPP,src,count,n);
      buffer[n]='\0';
      printf("%s %s: [%d] message received=\"%s\"\n",MYAPP,src,count,buffer);
    }
    else {
      printf("%s %s: [%d] %d bytes read! write-end of pipe has been closed!\n",MYAPP,src,count,n);
      more=0;
    }
    count++;
  }
}

int main ()
{
  int fds[2];
  pid_t pid;

  /* Create a pipe. File descriptors for the two ends of the pipe are
     placed in fds. */
  printf("%s %s: creating a pipe ...\n",MYAPP,src);
  pipe (fds);
  printf("%s %s: ... pipe created with fds[0]=%d (r), fds[1]=%d (w).\n",MYAPP,src,fds[0],fds[1]);

  /* Fork a child process. */
  printf("%s %s: forking to create a child process ...\n",MYAPP,src);
  pid = fork ();

  if (pid == (pid_t) 0) { // Child process
    src=CHILD_MSG;

    /* Close our copy of the write end of the pipe. */
    printf("%s %s: closing write-end of pipe (fds[1]=%d)\n",MYAPP,src,fds[1]);
    close (fds[1]);

    printf("%s %s: reading all the messages till EOF from read-end of pipe\n",
	   MYAPP,src);
    reader (fds[0]);

    printf("%s %s: closing read-end of pipe (fds[0]=%d)\n",MYAPP,src,fds[0]);
    close (fds[0]);
  }
  else { // Parent process
   src=PARENT_MSG;

    /* Close our copy of the read end of the pipe. */
    printf("%s %s: closing read-end of pipe (fds[0]=%d)\n",MYAPP,src,fds[0]);
    close (fds[0]);

    printf("%s %s: writing %d times the string \"%s\" into write-end of pipe\n",
	   MYAPP,src,NUM_TIMES,MESSAGE);
    writer (MESSAGE,NUM_TIMES,fds[1]);

    printf("%s %s: closing write-end of pipe (fds[1]=%d)\n",MYAPP,src,fds[1]);
    close (fds[1]);
  }
  return 0;
}
